@@ -9,11 +9,16 @@ struct ctx;
 struct output_list_node;
 
 typedef enum {
-    SCALE_LINEAR,
-    SCALE_NEAREST,
+    SCALE_FIT,
+    SCALE_COVER,
     SCALE_EXACT
 } scale_t;
 
+typedef enum {
+    SCALE_FILTER_LINEAR,
+    SCALE_FILTER_NEAREST,
+} scale_filter_t;
+
 typedef enum {
     BACKEND_AUTO,
     BACKEND_DMABUF,
@@ -27,17 +32,20 @@ typedef struct ctx_opt {
     bool invert_colors;
     bool freeze;
     bool has_region;
+    bool fullscreen;
     scale_t scaling;
+    scale_filter_t scaling_filter;
     backend_t backend;
     transform_t transform;
     region_t region;
     char * output;
+    char * fullscreen_output;
 } ctx_opt_t;
 
 void init_opt(struct ctx * ctx);
 void cleanup_opt(struct ctx * ctx);
 
-bool parse_scaling_opt(scale_t * scaling, const char * scaling_arg);
+bool parse_scaling_opt(scale_t * scaling, scale_filter_t * scaling_filter, const char * scaling_arg);
 bool parse_transform_opt(transform_t * transform, const char * transform_arg);
 bool parse_region_opt(region_t * region, char ** output, const char * region_arg);
 bool find_output_opt(struct ctx * ctx, struct output_list_node ** output_handle, region_t * region_handle);
