@@ -1,12 +1,16 @@
 #include <GLES2/gl2.h>
 #include <GLES2/gl2ext.h>
+#include <EGL/egl.h>
+#include <EGL/eglext.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include "context.h"
+#include "egl.h"
 #include "transform.h"
 #include "glsl_vertex_shader.h"
 #include "glsl_fragment_shader.h"
+#include "util.h"
 
 // --- buffers ---
 
@@ -110,7 +114,7 @@ void init_egl(ctx_t * ctx) {
     }
 
     // create egl surface
-    ctx->egl.surface = eglCreateWindowSurface(ctx->egl.display, ctx->egl.config, ctx->egl.window, NULL);
+    ctx->egl.surface = eglCreateWindowSurface(ctx->egl.display, ctx->egl.config, (EGLNativeWindowType)ctx->egl.window, NULL);
 
     // create egl context with support for OpenGL ES 2.0
     EGLint context_attribs[] = {
@@ -153,7 +157,7 @@ void init_egl(ctx_t * ctx) {
     // create texture and set scaling mode
     glGenTextures(1, &ctx->egl.texture);
     glBindTexture(GL_TEXTURE_2D, ctx->egl.texture);
-    if (ctx->opt.scaling == SCALE_LINEAR) {
+    if (ctx->opt.scaling_filter == SCALE_FILTER_LINEAR) {
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
     } else {
@@ -164,7 +168,7 @@ void init_egl(ctx_t * ctx) {
     // create freeze texture and set scaling mode
     glGenTextures(1, &ctx->egl.freeze_texture);
     glBindTexture(GL_TEXTURE_2D, ctx->egl.freeze_texture);
-    if (ctx->opt.scaling == SCALE_LINEAR) {
+    if (ctx->opt.scaling_filter == SCALE_FILTER_LINEAR) {
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
     } else {
@@ -272,8 +276,8 @@ void draw_texture(ctx_t *ctx) {
 void resize_viewport(ctx_t * ctx) {
     log_debug(ctx, "egl::resize_viewport(): resizing viewport\n");
 
-    uint32_t win_width = ctx->wl.scale * ctx->wl.width;
-    uint32_t win_height = ctx->wl.scale * ctx->wl.height;
+    uint32_t win_width = round(ctx->wl.width * ctx->wl.scale);
+    uint32_t win_height = round(ctx->wl.height * ctx->wl.scale);
     uint32_t tex_width = ctx->egl.width;
     uint32_t tex_height = ctx->egl.height;
     uint32_t view_width = win_width;
@@ -294,7 +298,10 @@ void resize_viewport(ctx_t * ctx) {
         };
         clamp_region = ctx->mirror.current_region;
 
-        region_scale(&clamp_region, ctx->mirror.current_target->scale);
+        // HACK: calculate effective output fractional scale
+        // wayland doesn't provide this information
+        double output_scale = (double)tex_width / ctx->mirror.current_target->width;
+        region_scale(&clamp_region, output_scale);
         region_clamp(&clamp_region, &output_region);
 
         tex_width = clamp_region.width;
@@ -304,28 +311,53 @@ void resize_viewport(ctx_t * ctx) {
     // rotate texture dimensions by user transform
     viewport_apply_transform(&tex_width, &tex_height, ctx->opt.transform);
 
-    // select biggest width or height that fits and preserves aspect ratio
-    float win_aspect = (float)win_width / win_height;
-    float tex_aspect = (float)tex_width / tex_height;
-    if (win_aspect > tex_aspect) {
-        view_width = view_height * tex_aspect;
-    } else if (win_aspect < tex_aspect) {
-        view_height = view_width / tex_aspect;
-    }
+    // calculate aspect ratio
+    double win_aspect = (double)win_width / win_height;
+    double tex_aspect = (double)tex_width / tex_height;
 
-    // find biggest fitting integer scale
-    if (ctx->opt.scaling == SCALE_EXACT) {
-        uint32_t width_scale = win_width / tex_width;
-        uint32_t height_scale = win_height / tex_height;
-        uint32_t scale = width_scale < height_scale ? width_scale : height_scale;
-        if (scale > 0) {
-            view_width = scale * tex_width;
-            view_height = scale * tex_height;
+    if (ctx->opt.scaling == SCALE_FIT) {
+        // select biggest width or height that fits and preserves aspect ratio
+        if (win_aspect > tex_aspect) {
+            view_width = view_height * tex_aspect;
+        } else if (win_aspect < tex_aspect) {
+            view_height = view_width / tex_aspect;
+        }
+    } else if (ctx->opt.scaling == SCALE_COVER) {
+        // select biggest width or height that covers and preserves aspect ratio
+        if (win_aspect < tex_aspect) {
+            view_width = view_height * tex_aspect;
+        } else if (win_aspect > tex_aspect) {
+            view_height = view_width / tex_aspect;
+        }
+    } else if (ctx->opt.scaling == SCALE_EXACT) {
+        // select biggest fitting integer scale
+        double width_scale = (double)win_width / tex_width;
+        double height_scale = (double)win_height / tex_height;
+        uint32_t upscale_factor = floorf(fminf(width_scale, height_scale));
+        uint32_t downscale_factor = ceilf(fmaxf(1 / width_scale, 1 / height_scale));
+
+        if (upscale_factor > 1) {
+            log_debug(ctx, "egl::resize_viewport(): upscaling by factor = %d\n", upscale_factor);
+            view_width = tex_width * upscale_factor;
+            view_height = tex_height * upscale_factor;
+        } else if (downscale_factor > 1) {
+            log_debug(ctx, "egl::resize_viewport(): downscaling by factor = %d\n", downscale_factor);
+            view_width = tex_width / downscale_factor;
+            view_height = tex_height / downscale_factor;
+        } else {
+            view_width = tex_width;
+            view_height = tex_height;
         }
     }
 
+    log_debug(ctx, "egl::resize_viewport(): win_width = %d, win_height = %d\n", win_width, win_height);
+    log_debug(ctx, "egl::resize_viewport(): view_width = %d, view_height = %d\n", view_width, view_height);
+
     // updating GL viewport
-    glViewport((win_width - view_width) / 2, (win_height - view_height) / 2, view_width, view_height);
+    log_debug(ctx, "egl::resize_viewport(): viewport %d, %d, %d, %d\n",
+        (int32_t)(win_width - view_width) / 2, (int32_t)(win_height - view_height) / 2, view_width, view_height
+    );
+    glViewport((int32_t)(win_width - view_width) / 2, (int32_t)(win_height - view_height) / 2, view_width, view_height);
 
     // recalculate texture transform
     mat3_t texture_transform;
@@ -354,10 +386,13 @@ void resize_viewport(ctx_t * ctx) {
 // --- resize_window ---
 
 void resize_window(ctx_t * ctx) {
-    log_debug(ctx, "egl::resize_window(): resizing EGL window\n");
+    uint32_t width = round(ctx->wl.width * ctx->wl.scale);
+    uint32_t height = round(ctx->wl.height * ctx->wl.scale);
+    log_debug(ctx, "egl::resize_window(): resizing EGL window to %dx%d\n", width, height);
 
     // resize window, then trigger viewport recalculation
-    wl_egl_window_resize(ctx->egl.window, ctx->wl.scale * ctx->wl.width, ctx->wl.scale * ctx->wl.height, 0, 0);
+    wl_egl_window_resize(ctx->egl.window, width, height, 0, 0);
+    wp_viewport_set_source(ctx->wl.viewport, 0, 0, wl_fixed_from_int(width), wl_fixed_from_int(height));
     resize_viewport(ctx);
 
     // redraw frame
@@ -380,7 +415,7 @@ void update_uniforms(ctx_t * ctx) {
 
     // set texture scaling mode
     glBindTexture(GL_TEXTURE_2D, ctx->egl.texture);
-    if (ctx->opt.scaling == SCALE_LINEAR) {
+    if (ctx->opt.scaling_filter == SCALE_FILTER_LINEAR) {
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
     } else {
@@ -389,7 +424,7 @@ void update_uniforms(ctx_t * ctx) {
     }
 
     glBindTexture(GL_TEXTURE_2D, ctx->egl.freeze_texture);
-    if (ctx->opt.scaling == SCALE_LINEAR) {
+    if (ctx->opt.scaling_filter == SCALE_FILTER_LINEAR) {
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
         glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
     } else {
@@ -407,6 +442,102 @@ void freeze_framebuffer(struct ctx * ctx) {
     glBindFramebuffer(GL_FRAMEBUFFER, 0);
 }
 
+// --- dmabuf_to_texture ---
+
+static const EGLAttrib fd_attribs[] = {
+    EGL_DMA_BUF_PLANE0_FD_EXT,
+    EGL_DMA_BUF_PLANE1_FD_EXT,
+    EGL_DMA_BUF_PLANE2_FD_EXT,
+    EGL_DMA_BUF_PLANE3_FD_EXT
+};
+_Static_assert(ARRAY_LENGTH(fd_attribs) == MAX_PLANES, "fd_attribs has incorrect length");
+
+static const EGLAttrib offset_attribs[] = {
+    EGL_DMA_BUF_PLANE0_OFFSET_EXT,
+    EGL_DMA_BUF_PLANE1_OFFSET_EXT,
+    EGL_DMA_BUF_PLANE2_OFFSET_EXT,
+    EGL_DMA_BUF_PLANE3_OFFSET_EXT
+};
+_Static_assert(ARRAY_LENGTH(offset_attribs) == MAX_PLANES, "offset_attribs has incorrect length");
+
+static const EGLAttrib stride_attribs[] = {
+    EGL_DMA_BUF_PLANE0_PITCH_EXT,
+    EGL_DMA_BUF_PLANE1_PITCH_EXT,
+    EGL_DMA_BUF_PLANE2_PITCH_EXT,
+    EGL_DMA_BUF_PLANE3_PITCH_EXT
+};
+_Static_assert(ARRAY_LENGTH(stride_attribs) == MAX_PLANES, "stride_attribs has incorrect length");
+
+static const EGLAttrib modifier_low_attribs[] = {
+    EGL_DMA_BUF_PLANE0_MODIFIER_LO_EXT,
+    EGL_DMA_BUF_PLANE1_MODIFIER_LO_EXT,
+    EGL_DMA_BUF_PLANE2_MODIFIER_LO_EXT,
+    EGL_DMA_BUF_PLANE3_MODIFIER_LO_EXT
+};
+_Static_assert(ARRAY_LENGTH(modifier_low_attribs) == MAX_PLANES, "modifier_low_attribs has incorrect length");
+
+static const EGLAttrib modifier_high_attribs[] = {
+    EGL_DMA_BUF_PLANE0_MODIFIER_HI_EXT,
+    EGL_DMA_BUF_PLANE1_MODIFIER_HI_EXT,
+    EGL_DMA_BUF_PLANE2_MODIFIER_HI_EXT,
+    EGL_DMA_BUF_PLANE3_MODIFIER_HI_EXT
+};
+_Static_assert(ARRAY_LENGTH(modifier_high_attribs) == MAX_PLANES, "modifier_high_attribs has incorrect length");
+
+bool dmabuf_to_texture(ctx_t * ctx, dmabuf_t * dmabuf) {
+    if (dmabuf->planes > MAX_PLANES) {
+        log_error("egl::dmabuf_to_texture(): too many planes, got %zd, can support at most %d\n", dmabuf->planes, MAX_PLANES);
+        return false;
+    }
+
+    int i = 0;
+    EGLAttrib * image_attribs = malloc((6 + 10 * dmabuf->planes + 1) * sizeof (EGLAttrib));
+    if (image_attribs == NULL) {
+        log_error("egl::dmabuf_to_texture(): failed to allocate EGL image attribs\n");
+        return false;
+    }
+
+    image_attribs[i++] = EGL_WIDTH;
+    image_attribs[i++] = dmabuf->width;
+    image_attribs[i++] = EGL_HEIGHT;
+    image_attribs[i++] = dmabuf->height;
+    image_attribs[i++] = EGL_LINUX_DRM_FOURCC_EXT;
+    image_attribs[i++] = dmabuf->drm_format;
+
+    for (size_t j = 0; j < dmabuf->planes; j++) {
+        image_attribs[i++] = fd_attribs[j];
+        image_attribs[i++] = dmabuf->fds[j];
+        image_attribs[i++] = offset_attribs[j];
+        image_attribs[i++] = dmabuf->offsets[j];
+        image_attribs[i++] = stride_attribs[j];
+        image_attribs[i++] = dmabuf->strides[j];
+        image_attribs[i++] = modifier_low_attribs[j];
+        image_attribs[i++] = (uint32_t)dmabuf->modifier;
+        image_attribs[i++] = modifier_high_attribs[j];
+        image_attribs[i++] = (uint32_t)(dmabuf->modifier >> 32);
+    }
+
+    image_attribs[i++] = EGL_NONE;
+
+    // create EGLImage from dmabuf with attribute array
+    EGLImage frame_image = eglCreateImage(ctx->egl.display, EGL_NO_CONTEXT, EGL_LINUX_DMA_BUF_EXT, NULL, image_attribs);
+    free(image_attribs);
+
+    if (frame_image == EGL_NO_IMAGE) {
+        log_error("egl::dmabuf_to_texture(): failed to create EGL image from dmabuf: error = %x\n", eglGetError());
+        return false;
+    }
+
+    // convert EGLImage to GL texture
+    glBindTexture(GL_TEXTURE_2D, ctx->egl.texture);
+    ctx->egl.glEGLImageTargetTexture2DOES(GL_TEXTURE_2D, frame_image);
+
+    // destroy temporary image
+    eglDestroyImage(ctx->egl.display, frame_image);
+
+    return true;
+}
+
 // --- cleanup_egl ---
 
 void cleanup_egl(ctx_t *ctx) {
