@@ -11,27 +11,37 @@ void init_opt(ctx_t * ctx) {
     ctx->opt.invert_colors = false;
     ctx->opt.freeze = false;
     ctx->opt.has_region = false;
-    ctx->opt.scaling = SCALE_LINEAR;
+    ctx->opt.fullscreen = false;
+    ctx->opt.scaling = SCALE_FIT;
+    ctx->opt.scaling_filter = SCALE_FILTER_LINEAR;
     ctx->opt.backend = BACKEND_AUTO;
     ctx->opt.transform = (transform_t){ .rotation = ROT_NORMAL, .flip_x = false, .flip_y = false };
     ctx->opt.region = (region_t){ .x = 0, .y = 0, .width = 0, .height = 0 };
     ctx->opt.output = NULL;
+    ctx->opt.fullscreen_output = NULL;
 }
 
 void cleanup_opt(ctx_t * ctx) {
     if (ctx->opt.output != NULL) free(ctx->opt.output);
+    if (ctx->opt.fullscreen_output != NULL) free(ctx ->opt.fullscreen_output);
 }
 
-bool parse_scaling_opt(scale_t * scaling, const char * scaling_arg) {
-    if (strcmp(scaling_arg, "l") == 0 || strcmp(scaling_arg, "linear") == 0) {
-        *scaling = SCALE_LINEAR;
+bool parse_scaling_opt(scale_t * scaling, scale_filter_t * scaling_filter, const char * scaling_arg) {
+    if (strcmp(scaling_arg, "f") == 0 || strcmp(scaling_arg, "fit") == 0) {
+        *scaling = SCALE_FIT;
         return true;
-    } else if (strcmp(scaling_arg, "n") == 0 || strcmp(scaling_arg, "nearest") == 0) {
-        *scaling = SCALE_NEAREST;
+    } else if (strcmp(scaling_arg, "c") == 0 || strcmp(scaling_arg, "cover") == 0) {
+        *scaling = SCALE_COVER;
         return true;
     } else if (strcmp(scaling_arg, "e") == 0 || strcmp(scaling_arg, "exact") == 0) {
         *scaling = SCALE_EXACT;
         return true;
+    } else if (strcmp(scaling_arg, "l") == 0 || strcmp(scaling_arg, "linear") == 0) {
+        *scaling_filter = SCALE_FILTER_LINEAR;
+        return true;
+    } else if (strcmp(scaling_arg, "n") == 0 || strcmp(scaling_arg, "nearest") == 0) {
+        *scaling_filter = SCALE_FILTER_NEAREST;
+        return true;
     } else {
         return false;
     }
@@ -330,25 +340,31 @@ void usage_opt(ctx_t * ctx) {
     printf("usage: wl-mirror [options] <output>\n");
     printf("\n");
     printf("options:\n");
-    printf("  -h,   --help             show this help\n");
-    printf("  -V,   --version          print version\n");
-    printf("  -v,   --verbose          enable debug logging\n");
-    printf("        --no-verbose       disable debug logging (default)\n");
-    printf("  -c,   --show-cursor      show the cursor on the mirrored screen (default)\n");
-    printf("        --no-show-cursor   don't show the cursor on the mirrored screen\n");
-    printf("  -i,   --invert-colors    invert colors in the mirrored screen\n");
-    printf("        --no-invert-colors don't invert colors in the mirrored screen (default)\n");
-    printf("  -f,   --freeze           freeze the current image on the screen\n");
-    printf("        --unfreeze         resume the screen capture after a freeze\n");
-    printf("        --toggle-freeze    toggle freeze state of screen capture\n");
-    printf("  -s l, --scaling linear   use linear scaling (default)\n");
-    printf("  -s n, --scaling nearest  use nearest neighbor scaling\n");
-    printf("  -s e, --scaling exact    only scale to exact multiples of the output size\n");
-    printf("  -b B  --backend B        use a specific backend for capturing the screen\n");
-    printf("  -t T, --transform T      apply custom transform T\n");
-    printf("  -r R, --region R         capture custom region R\n");
-    printf("        --no-region        capture the entire output (default)\n");
-    printf("  -S,   --stream           accept a stream of additional options on stdin\n");
+    printf("  -h,   --help                  show this help\n");
+    printf("  -V,   --version               print version\n");
+    printf("  -v,   --verbose               enable debug logging\n");
+    printf("        --no-verbose            disable debug logging (default)\n");
+    printf("  -c,   --show-cursor           show the cursor on the mirrored screen (default)\n");
+    printf("        --no-show-cursor        don't show the cursor on the mirrored screen\n");
+    printf("  -i,   --invert-colors         invert colors in the mirrored screen\n");
+    printf("        --no-invert-colors      don't invert colors in the mirrored screen (default)\n");
+    printf("  -f,   --freeze                freeze the current image on the screen\n");
+    printf("        --unfreeze              resume the screen capture after a freeze\n");
+    printf("        --toggle-freeze         toggle freeze state of screen capture\n");
+    printf("  -F,   --fullscreen            open wl-mirror as fullscreen\n");
+    printf("        --no-fullscreen         open wl-mirror as a window (default)\n");
+    printf("        --fullscreen-output O   open wl-mirror as fullscreen on output O\n");
+    printf("        --no-fullscreen-output  open wl-mirror as fullscreen on the current output (default)\n");
+    printf("  -s f, --scaling fit           scale to fit (default)\n");
+    printf("  -s c, --scaling cover         scale to cover, cropping if needed\n");
+    printf("  -s e, --scaling exact         only scale to exact multiples of the output size\n");
+    printf("  -s l, --scaling linear        use linear scaling (default)\n");
+    printf("  -s n, --scaling nearest       use nearest neighbor scaling\n");
+    printf("  -b B  --backend B             use a specific backend for capturing the screen\n");
+    printf("  -t T, --transform T           apply custom transform T\n");
+    printf("  -r R, --region R              capture custom region R\n");
+    printf("        --no-region             capture the entire output (default)\n");
+    printf("  -S,   --stream                accept a stream of additional options on stdin\n");
     printf("\n");
     printf("backends:\n");
     printf("  - auto        automatically try the backends in order and use the first that works (default)\n");
@@ -392,9 +408,11 @@ void version_opt(ctx_t * ctx) {
 void parse_opt(ctx_t * ctx, int argc, char ** argv) {
     bool is_cli_args = !ctx->opt.stream;
     bool was_frozen = ctx->opt.freeze;
+    bool was_fullscreen = ctx->opt.fullscreen;
     bool new_backend = false;
     bool new_region = false;
     bool new_output = false;
+    bool new_fullscreen_output = false;
     char * region_output = NULL;
     char * arg_output = NULL;
 
@@ -424,12 +442,32 @@ void parse_opt(ctx_t * ctx, int argc, char ** argv) {
             ctx->opt.freeze = false;
         } else if (strcmp(argv[0], "--toggle-freeze") == 0) {
             ctx->opt.freeze ^= 1;
+        } else if (strcmp(argv[0], "-F") == 0 || strcmp(argv[0], "--fullscreen") == 0) {
+            ctx->opt.fullscreen = true;
+        } else if (strcmp(argv[0], "--no-fullscreen") == 0) {
+            ctx->opt.fullscreen = false;
+        } else if (strcmp(argv[0], "--fullscreen-output") == 0) {
+            if (argc < 2) {
+                log_error("options::parse(): option %s requires an argument\n", argv[0]);
+                if (is_cli_args) exit_fail(ctx);
+            } else {
+                free(ctx->opt.fullscreen_output);
+                ctx->opt.fullscreen = true;
+                ctx->opt.fullscreen_output = strdup(argv[1]);
+                new_fullscreen_output = true;
+                argv++;
+                argc--;
+            }
+        } else if (strcmp(argv[0], "--no-fullscreen-output") == 0) {
+            free(ctx->opt.fullscreen_output);
+            ctx->opt.fullscreen_output = NULL;
+            new_fullscreen_output = true;
         } else if (strcmp(argv[0], "-s") == 0 || strcmp(argv[0], "--scaling") == 0) {
             if (argc < 2) {
                 log_error("options::parse(): option %s requires an argument\n", argv[0]);
                 if (is_cli_args) exit_fail(ctx);
             } else {
-                if (!parse_scaling_opt(&ctx->opt.scaling, argv[1])) {
+                if (!parse_scaling_opt(&ctx->opt.scaling, &ctx->opt.scaling_filter, argv[1])) {
                     log_error("options::parse(): invalid scaling mode %s\n", argv[1]);
                     if (is_cli_args) exit_fail(ctx);
                 }
@@ -548,11 +586,25 @@ void parse_opt(ctx_t * ctx, int argc, char ** argv) {
         usage_opt(ctx);
     }
 
+    if (
+        ctx->opt.output != NULL && ctx->opt.fullscreen_output != NULL &&
+        strcmp(ctx->opt.output, ctx->opt.fullscreen_output) == 0
+    ) {
+        log_error("options::parse(): fullscreen_output cannot be same as the output to be mirrored");
+        exit_fail(ctx);
+    }
+
     if (argc > 1) {
         log_error("options::parse(): unexpected trailing arguments after output name\n");
         if (is_cli_args) exit_fail(ctx);
     }
 
+    if (!is_cli_args && ctx->opt.fullscreen && (!was_fullscreen || new_fullscreen_output)) {
+        set_window_fullscreen(ctx);
+    } else if (!is_cli_args && !ctx->opt.fullscreen && was_fullscreen) {
+        unset_window_fullscreen(ctx);
+    }
+
     output_list_node_t * target_output = NULL;
     region_t target_region = (region_t){ .x = 0, .y = 0, .width = 0, .height = 0 };
     if (!is_cli_args && find_output_opt(ctx, &target_output, &target_region)) {
