@@ -94,13 +94,7 @@ static void on_output_scale(
         // update buffer scale only if this is the current output
         if (ctx->wl.current_output != NULL && ctx->wl.current_output->output == output) {
             log_debug(ctx, "wayland::on_output_scale(): updating window scale\n");
-            ctx->wl.scale = scale;
-            wl_surface_set_buffer_scale(ctx->wl.surface, scale);
-
-            // resize egl window to reflect new scale
-            if (ctx->egl.initialized) {
-                resize_window(ctx);
-            }
+            update_window_scale(ctx, scale, false);
         }
     }
 }
@@ -237,6 +231,26 @@ static void on_registry_add(
         ctx->wl.seat = (struct wl_seat *)wl_registry_bind(
             registry, id, &wl_seat_interface, 1
         );
+    } else if (strcmp(interface, wp_viewporter_interface.name) == 0) {
+        if (ctx->wl.viewporter != NULL) {
+            log_error("wayland::on_registry_add(): duplicate wp_viewporter\n");
+            exit_fail(ctx);
+        }
+
+        // bind wp_viewporter object
+        ctx->wl.viewporter = (struct wp_viewporter *)wl_registry_bind(
+            registry, id, &wp_viewporter_interface, 1
+        );
+    } else if (strcmp(interface, wp_fractional_scale_manager_v1_interface.name) == 0) {
+        if (ctx->wl.fractional_scale_manager != NULL) {
+            log_error("wayland::on_registry_add(): duplicate wp_fractional_scale_manager\n");
+            exit_fail(ctx);
+        }
+
+        // bind wp_fractional_scale_manager_v1 object
+        ctx->wl.fractional_scale_manager = (struct wp_fractional_scale_manager_v1 *)wl_registry_bind(
+            registry, id, &wp_fractional_scale_manager_v1_interface, 1
+        );
     } else if (strcmp(interface, xdg_wm_base_interface.name) == 0) {
         if (ctx->wl.wm_base != NULL) {
             log_error("wayland::on_registry_add(): duplicate wm_base\n");
@@ -365,19 +379,25 @@ static void on_registry_remove(
     // ensure protocols we need aren't removed
     // remove removed outputs from the output list
     if (id == ctx->wl.compositor_id) {
-        log_error("wayland::on_registry_remove(): compositor disapperared\n");
+        log_error("wayland::on_registry_remove(): compositor disappeared\n");
         exit_fail(ctx);
     } else if (id == ctx->wl.seat_id) {
-        log_error("wayland::on_registry_remove(): seat disapperared\n");
+        log_error("wayland::on_registry_remove(): seat disappeared\n");
+        exit_fail(ctx);
+    } else if (id == ctx->wl.viewporter_id) {
+        log_error("wayland::on_registry_remove(): viewporter disappeared\n");
+        exit_fail(ctx);
+    } else if (id == ctx->wl.fractional_scale_manager_id) {
+        log_error("wayland::on_registry_remove(): fractional_scale_manager disappeared\n");
         exit_fail(ctx);
     } else if (id == ctx->wl.wm_base_id) {
-        log_error("wayland::on_registry_remove(): wm_base disapperared\n");
+        log_error("wayland::on_registry_remove(): wm_base disappeared\n");
         exit_fail(ctx);
     } else if (id == ctx->wl.output_manager_id) {
-        log_error("wayland::on_registry_remove(): output_manager disapperared\n");
+        log_error("wayland::on_registry_remove(): output_manager disappeared\n");
         exit_fail(ctx);
     } else if (id == ctx->wl.dmabuf_manager_id) {
-        log_error("wayland::on_registry_remove(): dmabuf_manager disapperared\n");
+        log_error("wayland::on_registry_remove(): dmabuf_manager disappeared\n");
         exit_fail(ctx);
     } else {
         output_list_node_t ** link = &ctx->wl.outputs;
@@ -454,25 +474,28 @@ static void on_surface_enter(
 
     // verify an output was found
     if (node == NULL) {
-        log_error("wayland::on_surface_enter(): entered nonexistent output\n");
-        exit_fail(ctx);
+        // when multiple registries exist and different parts of the application
+        // bind separately to the same wl_output, on_surface_enter will be called
+        // multiple times.
+        //
+        // only one of those calls will give us an output in our list of outputs.
+        // we should ignore all other calls, instead of bailing completely
+        log_debug(ctx, "wayland::on_surface_enter(): entered output not in output list, possibly bound from another registry?\n");
+        return;
     }
 
     // update current output
+    bool first_output = ctx->wl.current_output == NULL;
     ctx->wl.current_output = node;
 
-    // update scale only if changed
-    if (node->scale != ctx->wl.scale) {
-        log_debug(ctx, "wayland::on_surface_enter(): updating window scale\n");
-        ctx->wl.scale = node->scale;
-        wl_surface_set_buffer_scale(ctx->wl.surface, node->scale);
-
-        // resize egl window to reflect new scale
-        if (ctx->egl.initialized) {
-            resize_window(ctx);
-        }
+    // set window fullscreen now if no specific output requested
+    if (first_output && ctx->opt.fullscreen) {
+        set_window_fullscreen(ctx);
     }
 
+    log_debug(ctx, "wayland::on_surface_enter(): updating window scale\n");
+    update_window_scale(ctx, node->scale, false);
+
     (void)surface;
 }
 
@@ -505,7 +528,7 @@ static void on_surface_configure_finished(ctx_t * ctx) {
 
 // --- xdg_surface event handlers ---
 
-static void xdg_on_surface_configure(
+static void on_xdg_surface_configure(
     void * data, struct xdg_surface * xdg_surface, uint32_t serial
 ) {
     ctx_t * ctx = (ctx_t *)data;
@@ -523,7 +546,7 @@ static void xdg_on_surface_configure(
 }
 
 static const struct xdg_surface_listener xdg_surface_listener = {
-    .configure = xdg_on_surface_configure,
+    .configure = on_xdg_surface_configure,
 };
 
 // --- xdg_toplevel event handlers ---
@@ -538,12 +561,30 @@ static void on_xdg_toplevel_configure(
     if (width == 0) width = 100;
     if (height == 0) height = 100;
 
+    // check fullscreen state
+    bool is_fullscreen = false;
+    uint32_t * state;
+    wl_array_for_each(state, states) {
+        if (*state == XDG_TOPLEVEL_STATE_FULLSCREEN) {
+            is_fullscreen = true;
+        }
+    }
+
+    // reset fullscreen option if the window is not fullscreen
+    // but only if have already had the chance to fullscreen on the current output
+    if (is_fullscreen != ctx->opt.fullscreen && ctx->wl.current_output != NULL) {
+        ctx->opt.fullscreen = is_fullscreen;
+    }
+
     // update size only if changed
     if (ctx->wl.width != (uint32_t)width || ctx->wl.height != (uint32_t)height) {
         log_debug(ctx, "wayland::on_xdg_toplevel_configure(): window resized to %dx%d\n", width, height);
         ctx->wl.width = width;
         ctx->wl.height = height;
 
+        // resize viewport
+        wp_viewport_set_destination(ctx->wl.viewport, width, height);
+
         // resize window to reflect new surface size
         if (ctx->egl.initialized) {
             resize_window(ctx);
@@ -576,6 +617,52 @@ static const struct xdg_toplevel_listener xdg_toplevel_listener = {
     .close = on_xdg_toplevel_close
 };
 
+static void on_loop_event(ctx_t * ctx) {
+    if (wl_display_dispatch(ctx->wl.display) == -1) {
+        ctx->wl.closing = true;
+    }
+}
+
+static void on_loop_each(ctx_t * ctx) {
+    wl_display_flush(ctx->wl.display);
+}
+
+// --- fractional scale event handlers ---
+
+static void on_fractional_scale_preferred_scale(void * data, struct wp_fractional_scale_v1 * fractional_scale, uint32_t scale_times_120) {
+    ctx_t * ctx = (ctx_t *)data;
+    double scale = scale_times_120 / 120.0;
+    log_debug(ctx, "wayland::on_fractional_scale_preferred_scale(): scale = %.4f\n", scale);
+
+    // fractionally scaled surfaces have buffer scale of 1
+    update_window_scale(ctx, scale, true);
+
+    (void)fractional_scale;
+}
+
+static const struct wp_fractional_scale_v1_listener fractional_scale_listener = {
+    .preferred_scale = on_fractional_scale_preferred_scale
+};
+
+// --- find_output ---
+
+bool find_wl_output(ctx_t * ctx, char * output_name, struct wl_output ** output) {
+    bool found = false;
+    output_list_node_t * cur = ctx->wl.outputs;
+    while (cur != NULL) {
+        if (cur->name != NULL && strcmp(cur->name, output_name) == 0) {
+            output_name = cur->name;
+            break;
+        }
+        cur = cur->next;
+    }
+    if (cur != NULL) {
+        found = true;
+        *output = cur->output;
+    }
+    return found;
+}
+
 // --- init_wl ---
 
 void init_wl(ctx_t * ctx) {
@@ -587,6 +674,10 @@ void init_wl(ctx_t * ctx) {
     ctx->wl.compositor_id = 0;
     ctx->wl.seat = NULL;
     ctx->wl.seat_id = 0;
+    ctx->wl.viewporter = NULL;
+    ctx->wl.viewporter_id = 0;
+    ctx->wl.fractional_scale_manager = NULL;
+    ctx->wl.fractional_scale_manager_id = 0;
     ctx->wl.wm_base = NULL;
     ctx->wl.wm_base_id = 0;
     ctx->wl.output_manager = NULL;
@@ -602,13 +693,20 @@ void init_wl(ctx_t * ctx) {
     ctx->wl.outputs = NULL;
 
     ctx->wl.surface = NULL;
+    ctx->wl.viewport = NULL;
+    ctx->wl.fractional_scale = NULL;
     ctx->wl.xdg_surface = NULL;
     ctx->wl.xdg_toplevel = NULL;
 
     ctx->wl.current_output = NULL;
     ctx->wl.width = 0;
     ctx->wl.height = 0;
-    ctx->wl.scale = 1;
+    ctx->wl.scale = 1.0;
+
+    ctx->wl.event_handler.fd = -1;
+    ctx->wl.event_handler.events = EPOLLIN;
+    ctx->wl.event_handler.on_event = on_loop_event;
+    ctx->wl.event_handler.on_each = on_loop_each;
 
     ctx->wl.last_surface_serial = 0;
     ctx->wl.xdg_surface_configured = false;
@@ -624,6 +722,10 @@ void init_wl(ctx_t * ctx) {
         exit_fail(ctx);
     }
 
+    // register event loop
+    ctx->wl.event_handler.fd = wl_display_get_fd(ctx->wl.display);
+    event_add_fd(ctx, &ctx->wl.event_handler);
+
     // get registry handle
     ctx->wl.registry = wl_display_get_registry(ctx->wl.display);
     if (ctx->wl.registry == NULL) {
@@ -645,6 +747,9 @@ void init_wl(ctx_t * ctx) {
     if (ctx->wl.compositor == NULL) {
         log_error("wayland::init(): compositor missing\n");
         exit_fail(ctx);
+    } else if (ctx->wl.viewporter == NULL) {
+        log_error("wayland::init(): viewporter missing\n");
+        exit_fail(ctx);
     } else if (ctx->wl.wm_base == NULL) {
         log_error("wayland::init(): wm_base missing\n");
         exit_fail(ctx);
@@ -669,6 +774,20 @@ void init_wl(ctx_t * ctx) {
     // - for leave event
     wl_surface_add_listener(ctx->wl.surface, &surface_listener, (void *)ctx);
 
+    // create viewport
+    ctx->wl.viewport = wp_viewporter_get_viewport(ctx->wl.viewporter, ctx->wl.surface);
+    if (ctx->wl.viewport == NULL) {
+        log_error("wayland::init(): failed to create viewport\n");
+        exit_fail(ctx);
+    }
+
+    // create fractional scale if supported
+    if (ctx->wl.fractional_scale_manager != NULL) {
+        ctx->wl.fractional_scale = wp_fractional_scale_manager_v1_get_fractional_scale(ctx->wl.fractional_scale_manager, ctx->wl.surface);
+
+        wp_fractional_scale_v1_add_listener(ctx->wl.fractional_scale, &fractional_scale_listener, (void *)ctx);
+    }
+
     // create xdg surface
     ctx->wl.xdg_surface = xdg_wm_base_get_xdg_surface(ctx->wl.wm_base, ctx->wl.surface);
     if (ctx->wl.xdg_surface == NULL) {
@@ -700,10 +819,14 @@ void init_wl(ctx_t * ctx) {
     wl_surface_commit(ctx->wl.surface);
 
     // wait for events
-    // - expecing surface configure event
+    // - expecting surface configure event
     // - expecting xdg toplevel configure event
     wl_display_roundtrip(ctx->wl.display);
 
+    // set fullscreen on xdg_toplevel
+    if (ctx->opt.fullscreen && ctx->opt.fullscreen_output != NULL) {
+        set_window_fullscreen(ctx);
+    }
     // check if surface is configured
     // - expecting surface to be configured at this point
     if (!ctx->wl.configured) {
@@ -712,6 +835,50 @@ void init_wl(ctx_t * ctx) {
     }
 }
 
+// --- set_window_title ---
+
+void set_window_title(ctx_t * ctx, const char * title) {
+    xdg_toplevel_set_title(ctx->wl.xdg_toplevel, title);
+}
+
+// --- set_window_fullscreen ---
+
+void set_window_fullscreen(ctx_t * ctx) {
+    struct wl_output * output = NULL;
+    if (ctx->opt.fullscreen_output == NULL) {
+        output = ctx->wl.current_output->output;
+    } else if (!find_wl_output(ctx, ctx->opt.fullscreen_output, &output)) {
+        log_error("wayland::init(): output %s not found\n", ctx->opt.fullscreen_output);
+        exit_fail(ctx);
+    }
+
+    xdg_toplevel_set_fullscreen(ctx->wl.xdg_toplevel, output);
+}
+
+void unset_window_fullscreen(ctx_t * ctx) {
+    xdg_toplevel_unset_fullscreen(ctx->wl.xdg_toplevel);
+}
+
+// --- update_window_scale
+
+void update_window_scale(ctx_t * ctx, double scale, bool is_fractional) {
+    bool resize = false;
+
+    // don't update scale from other sources if fractional scaling supported
+    if (ctx->wl.fractional_scale != NULL && !is_fractional) return;
+
+    if (ctx->wl.scale != scale) {
+        log_debug(ctx, "wayland::update_window_scale(): setting window scale to %.4f\n", scale);
+        ctx->wl.scale = scale;
+        resize = true;
+    }
+
+    // resize egl window to reflect new scale
+    if (resize && ctx->egl.initialized) {
+        resize_window(ctx);
+    }
+}
+
 // --- cleanup_wl ---
 
 void cleanup_wl(ctx_t *ctx) {
@@ -719,6 +886,9 @@ void cleanup_wl(ctx_t *ctx) {
 
     log_debug(ctx, "wayland::cleanup(): destroying wayland objects\n");
 
+    // deregister event handler
+    event_remove_fd(ctx, &ctx->wl.event_handler);
+
     // free every output in output list
     output_list_node_t * cur = ctx->wl.outputs;
     output_list_node_t * prev = NULL;
@@ -739,9 +909,13 @@ void cleanup_wl(ctx_t *ctx) {
     if (ctx->wl.shm != NULL) wl_shm_destroy(ctx->wl.shm);
     if (ctx->wl.xdg_toplevel != NULL) xdg_toplevel_destroy(ctx->wl.xdg_toplevel);
     if (ctx->wl.xdg_surface != NULL) xdg_surface_destroy(ctx->wl.xdg_surface);
+    if (ctx->wl.fractional_scale != NULL) wp_fractional_scale_v1_destroy(ctx->wl.fractional_scale);
+    if (ctx->wl.viewport != NULL) wp_viewport_destroy(ctx->wl.viewport);
     if (ctx->wl.surface != NULL) wl_surface_destroy(ctx->wl.surface);
     if (ctx->wl.output_manager != NULL) zxdg_output_manager_v1_destroy(ctx->wl.output_manager);
     if (ctx->wl.wm_base != NULL) xdg_wm_base_destroy(ctx->wl.wm_base);
+    if (ctx->wl.fractional_scale_manager != NULL) wp_fractional_scale_manager_v1_destroy(ctx->wl.fractional_scale_manager);
+    if (ctx->wl.viewporter != NULL) wp_viewporter_destroy(ctx->wl.viewporter);
     if (ctx->wl.seat != NULL) wl_seat_destroy(ctx->wl.seat);
     if (ctx->wl.compositor != NULL) wl_compositor_destroy(ctx->wl.compositor);
     if (ctx->wl.registry != NULL) wl_registry_destroy(ctx->wl.registry);
